library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use work.funciones.all;

entity cordic is
  generic(
    N: natural               := 16; -- Ancho de la palabra
    ITER: natural            := 16; -- Numero de iteraciones
    FRACTIONAL_BITS: natural :=  8  -- Cantidad de bits fraccionarios
  );
  port(
    clk : in std_logic;
    rst : in std_logic;
    en_i : in std_logic;
    x_i  : in std_logic_vector(N-1 downto 0);
    y_i  : in std_logic_vector(N-1 downto 0);
    z_i  : in std_logic_vector(N-1 downto 0);
    dv_o : out std_logic;
    x_o  : out std_logic_vector(N-1 downto 0);
    y_o  : out std_logic_vector(N-1 downto 0);
    z_o  : out std_logic_vector(N-1 downto 0)
  );
end entity;

architecture structural of cordic is

  component cordic_iter is
    generic(
      N     : natural := 8;  --Ancho de la palabra
      SHIFT : natural := 1); --Desplazamiento
    port(
      clk   : in std_logic;
      rst   : in std_logic;
      en_i  : in std_logic;
      xi    : in std_logic_vector (N-1 downto 0);
      yi    : in std_logic_vector (N-1 downto 0);
      zi    : in std_logic_vector (N-1 downto 0);
      ci    : in std_logic_vector (N-1 downto 0);
      dv_o  : out std_logic;
      xip1  : out std_logic_vector (N-1 downto 0);
      yip1  : out std_logic_vector (N-1 downto 0);
      zip1  : out std_logic_vector (N-1 downto 0)
    );
  end component;

  type handShakeVector is array(ITER downto 0) of std_logic;
  signal en, dv : handShakeVector;
  type ConnectVector is array(ITER downto 0) of std_logic_vector(N-1 downto 0);
  signal wirex, wirey, wirez, wireLUT : ConnectVector;

  signal atanLUT: lut_t(ITER downto 0) := cordic_lut(N, FRACTIONAL_BITS, ITER);
  constant growth_coeff: real := correction_coeff(N, FRACTIONAL_BITS, ITER);
  constant growth_coeff_fixpoint: signed := to_signed(natural(growth_coeff * real(2 ** FRACTIONAL_BITS)), N);

  signal normalized_x, normalized_y: signed(2*N - 1 downto 0);

  signal is_2nd_or_third_quadrant: std_logic;
  signal corrected_x_i: std_logic_vector(N-1 downto 0);
  signal corrected_y_i: std_logic_vector(N-1 downto 0);
  signal corrected_z_i: std_logic_vector(N-1 downto 0);
begin

corrected_x_i <= std_logic_vector(-signed(x_i));
corrected_y_i <= std_logic_vector(-signed(y_i));
corrected_z_i <= std_logic_vector(signed(z_i) - signed(to_fixpoint(180, N, FRACTIONAL_BITS)));

en(0) <= en_i;
is_2nd_or_third_quadrant <= x_i(N-1);
wirex(0) <= x_i when is_2nd_or_third_quadrant = '0' else corrected_x_i;
wirey(0) <= y_i when is_2nd_or_third_quadrant = '0' else corrected_y_i;
wirez(0) <= z_i when is_2nd_or_third_quadrant = '0' else corrected_z_i;

CONNECTION_INSTANCE: for j in 0 to ITER-1 generate
  begin
    wireLUT(j) <= std_logic_vector(to_unsigned(atanLUT(j),N));

    ITERATION: cordic_iter
      generic map(N,j)
      port map(
        clk => clk,
        rst => rst,
        en_i  => en(j),
        xi    => wirex(j),
        yi    => wirey(j),
        zi    => wirez(j),
        ci   => wireLUT(j),
        dv_o => dv(j),
        xip1 => wirex(j+1),
        yip1 => wirey(j+1),
        zip1 => wirez(j+1)
        );

    en(j+1) <= dv(j);

  end generate;

dv_o <= dv(ITER-1);

normalized_x <= shift_right(signed(wirex(ITER)) * growth_coeff_fixpoint, FRACTIONAL_BITS);
normalized_y <= shift_right(signed(wirey(ITER)) * growth_coeff_fixpoint, FRACTIONAL_BITS);

x_o <= std_logic_vector(resize(normalized_x, N));
y_o <= std_logic_vector(resize(normalized_y, N));
z_o <= wirez(ITER);

end architecture;
