library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity adsr is
	port (
		clk:    in std_logic;
		sample: in std_logic_vector(15 downto 0);
		volume: in std_logic_vector(15 downto 0);
		note_on: in std_logic;
		note_off: in std_logic;

        attack_increment: in std_logic_vector(15 downto 0);
        decay_decrement: in std_logic_vector(15 downto 0);
        sustain_level: in std_logic_vector(15 downto 0);
        release_decrement: in std_logic_vector(15 downto 0);

		is_note_playing: out std_logic;
		rescaled_sample: out std_logic_vector(15 downto 0)
	);
end entity;

architecture structural of adsr is
	signal updated_sample: std_logic_vector(16+17+17-1 downto 0);

    type adsr_state_t is (attacking, decaying, sustaining, releasing, free);
    signal adsr_state: adsr_state_t;
    signal adsr_volume: unsigned(15 downto 0);
    signal next_adsr_volume: unsigned(15 downto 0);
    signal next_adsr_volume_when_attack: unsigned(15 downto 0);
    signal next_adsr_volume_when_decay: unsigned(15 downto 0);
    signal next_adsr_volume_when_release: unsigned(15 downto 0);

    constant min_u16: unsigned(15 downto 0) := "0000000000000000";
    constant max_u16: unsigned(15 downto 0) := "1111111111111111";
begin

    updated_sample <= (others => '0') when volume = "00000000" or adsr_state = free else
        std_logic_vector(signed(sample) * signed('0' & adsr_volume) * signed('0' & volume));

    -- A note is playing if we are not on an idle state
    is_note_playing <= '0' when adsr_state = free else '1';

    next_adsr_volume_when_attack <= max_u16 when adsr_volume + unsigned(attack_increment) < adsr_volume
        else adsr_volume + unsigned(attack_increment);

    next_adsr_volume_when_decay <= unsigned(sustain_level) when unsigned(sustain_level) > adsr_volume
        else adsr_volume - unsigned(decay_decrement);

    next_adsr_volume_when_release <= min_u16 when adsr_volume - unsigned(release_decrement) > adsr_volume
        else adsr_volume - unsigned(release_decrement);

    next_adsr_volume <= next_adsr_volume_when_attack when adsr_state = attacking else
        next_adsr_volume_when_decay when adsr_state = decaying else
        next_adsr_volume_when_release when adsr_state = releasing else
        min_u16 when adsr_state = free else
        unsigned(sustain_level);

	update_output: process(clk)
	begin
		if (rising_edge(clk)) then
			rescaled_sample <= updated_sample(16+16+16-1 downto 16+16);

            case std_logic_vector'(note_on, note_off) is
                when "10" =>
                    adsr_state <= attacking;
                    adsr_volume <= min_u16;
                when "11" =>
                    adsr_state <= attacking;
                    adsr_volume <= min_u16;
                when "01" =>
                    adsr_state <= releasing;
                    adsr_volume <= unsigned(sustain_level);
                when others =>
                    case adsr_state is
                        when free =>
                            -- Keep idling on zero
                            adsr_volume <= min_u16;
                        when sustaining =>
                            -- Dont advance once we reached sustain
                            adsr_volume <= unsigned(sustain_level);
                        when others =>
                            -- Modify state accordingly
                            if adsr_state = attacking and adsr_volume = max_u16 then
                                adsr_state <= decaying;
                            elsif adsr_state = decaying and adsr_volume = unsigned(sustain_level) then
                                adsr_state <= sustaining;
                            elsif adsr_state = releasing and adsr_volume = min_u16 then
                                adsr_state <= free;
                            else
                                adsr_state <= adsr_state;
                            end if;
                            -- Update volume
                            adsr_volume <= next_adsr_volume;
                    end case;
            end case;
		end if;
	end process;
end architecture;
